
/* Chrysalide - Outil d'analyse de fichiers binaires
 * format.c - équivalent Python du fichier "plugins/elf/format.c"
 *
 * Copyright (C) 2013-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "format.h"


#include <pygobject.h>


#include <i18n.h>
#include <format/executable.h>
#include <analysis/content.h>
#include <plugins/pychrysa/helpers.h>
#include <plugins/pychrysa/analysis/content.h>
#include <plugins/pychrysa/format/executable.h>


#include "constants.h"
#include "dynamic.h"
#include "../format.h"



/* Crée un nouvel objet Python de type 'ElfFormat'. */
static PyObject *py_elf_format_new(PyTypeObject *, PyObject *, PyObject *);



/******************************************************************************
*                                                                             *
*  Paramètres  : type = type de l'objet à instancier.                         *
*                args = arguments fournis à l'appel.                          *
*                kwds = arguments de type key=val fournis.                    *
*                                                                             *
*  Description : Crée un nouvel objet Python de type 'ElfFormat'.             *
*                                                                             *
*  Retour      : Instance Python mise en place.                               *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_elf_format_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    PyObject *result;                       /* Instance à retourner        */
    PyObject *content_obj;                  /* Objet pour le contenu       */
    PyObject *parent_obj;                   /* Objet pour le parent        */
    PyObject *status_obj;                   /* Objet pour la progression   */
    int ret;                                /* Bilan de lecture des args.  */
    GBinContent *content;                   /* Instance GLib du contenu    */
    GExeFormat *parent;                     /* Instance GLib du parent     */
    GtkStatusStack *status;                 /* Instance GTK de suivi       */
    GBinFormat *format;                     /* Création GLib à transmettre */

    parent_obj = Py_None;
    status_obj = Py_None;

    ret = PyArg_ParseTuple(args, "O|OO", &content_obj, &parent_obj, &status_obj);
    if (!ret) return NULL;

    ret = PyObject_IsInstance(content_obj, (PyObject *)get_python_binary_content_type());
    if (!ret)
    {
        PyErr_SetString(PyExc_TypeError, _("The first argument must be an instance of BinContent."));
        return NULL;
    }

    content = G_BIN_CONTENT(pygobject_get(content_obj));

    if (parent_obj == Py_None)
        parent = NULL;

    else
    {
        ret = PyObject_IsInstance(parent_obj, (PyObject *)get_python_executable_format_type());
        if (!ret)
        {
            PyErr_SetString(PyExc_TypeError, _("The second argument must be a container format or None."));
            return NULL;
        }

        parent = G_EXE_FORMAT(pygobject_get(parent_obj));

    }

    if (status_obj == Py_None)
        status = NULL;

    else
    {
        ret = PyObject_IsInstance(status_obj, (PyObject *)get_python_binary_content_type());
        if (!ret)
        {
            PyErr_SetString(PyExc_TypeError, _("The third argument must be a status bar object or None."));
            return NULL;
        }

        status = GTK_STATUS_STACK(pygobject_get(status_obj));

    }

    format = g_elf_format_new(content, parent, status);
    if (format == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, _("Unable to load the ELF format."));
        return NULL;
    }

    result = pygobject_new(G_OBJECT(format));

    g_object_unref(format);

    return (PyObject *)result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_elf_format_type(void)
{
    static PyMethodDef py_elf_format_methods[] = {
        { NULL }
    };

    static PyGetSetDef py_elf_format_getseters[] = {
        {
            "needed", py_elf_format_get_needed, NULL,
            "Provide the list of requiered shared objects.", NULL
        },
        { NULL }
    };

    static PyTypeObject py_elf_format_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.format.elf.ElfFormat",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT,

        .tp_doc         = "PyChrysalide Elf format",

        .tp_methods     = py_elf_format_methods,
        .tp_getset      = py_elf_format_getseters,
        .tp_new         = (newfunc)py_elf_format_new

    };

    return &py_elf_format_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.format.elf.ElfFormat'. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool register_python_elf_format(PyObject *module)
{
    PyTypeObject *py_elf_format_type;       /* Type Python 'ElfFormat'     */
    PyObject *dict;                         /* Dictionnaire du module      */

    py_elf_format_type = get_python_elf_format_type();

    dict = PyModule_GetDict(module);

    if (!register_class_for_pygobject(dict, G_TYPE_ELF_FORMAT,
                                      py_elf_format_type, get_python_executable_format_type()))
        return false;

    if (!define_python_binary_format_constants(py_elf_format_type))
        return false;

    return true;

}
